/*
** iplog_pcap.c - iplog pcap management routines.
** Copyright (C) 1999 Odin <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
*/

#include <config.h>

#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#ifdef HAVE_PCAP_H
#	include <pcap.h>
#elif defined HAVE_PCAP_PCAP_H
#	include <pcap/pcap.h>
#else
#	error "Cannot find pcap header files."
#endif
#include <syslog.h>
#include <pthread.h>

#include <assert.h>

#include <iplog.h>
#include <iplog_pcap.h>
#include <iplog_options.h>

extern u_long flags;

static int open_pcap_device(struct iff_data *idata, struct pcap_data *pdata);
static int get_if_namesaddr(struct iff_data **int_list, const u_char **vif);
static int get_pcap_datalink(pcap_t *pd);
int in_cksum(u_short *addr, int len);

static u_char errbuf[2048];
static struct frag_list frags[FRAG_TSIZE];

static int get_if_namesaddr(struct iff_data **int_list, const u_char **vif) {
	int sock;
#ifdef HAVE_SOCKADDR_SA_LEN
	u_int i;
#endif
	u_int z;
	struct ifconf ifc;
	struct ifreq buf[16], *ifr, *last, *next, ifrf;
	struct iff_data data;
	struct sockaddr_in *sin;
	u_char *p, *nlast = "";

	memset(&data, 0, sizeof(data));

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock == -1)
		return (-1);

	ifc.ifc_len = sizeof(buf);
	ifc.ifc_buf = (caddr_t) buf;
	
	if (ioctl(sock, SIOCGIFCONF, &ifc) == -1 ||
		(u_long) ifc.ifc_len < sizeof(struct ifreq))
	{
		close(sock);
		return (-1);
	}

	ifr = buf;
	last = (struct ifreq *) ((char *) buf + ifc.ifc_len);
	for (; ifr < last ; ifr = next) {
#ifdef HAVE_SOCKADDR_SA_LEN
		i = ifr->ifr_addr.sa_len + sizeof(ifr->ifr_name);
		if (i < sizeof(struct ifreq))
			next = ifr + 1;
		else
			next = (struct ifreq *) ((char *) ifr + i);
#else
		next = ifr + 1;
#endif
		xstrncpy(ifrf.ifr_name, ifr->ifr_name, sizeof(ifrf.ifr_name));

		if (ioctl(sock, SIOCGIFFLAGS, &ifrf) == -1)
			continue;
		if ((ifrf.ifr_flags & IFF_UP) && ifr->ifr_addr.sa_family == AF_INET
			&& !LOOPBACK(ifr->ifr_name))
		{
			if (vif != NULL) {
				bool found = false;
				u_char *a = xstrdup(ifr->ifr_name);
				u_char *b = strchr(a, ':');
				if (b != NULL)
					*b = '\0';

				for (z = 0 ; vif[z] != NULL ;) {
					if (strchr(vif[z], ':') != NULL)
						b = ifr->ifr_name;
					else
						b = a;

					if (!strcmp(vif[z++], b)) {
						found = true;
						break;
					}
				}

				free(a);
				if (found == false)
					continue;
			}

			p = strchr(ifr->ifr_name, ':');
			if (p != NULL)
				*p = '\0';

			if (strcmp(nlast, ifr->ifr_name)) {
				if (strcmp(nlast, ""))
					list_append(&data, (void **) int_list, sizeof(data));
				nlast = ifr->ifr_name;
				memset(&data, 0, sizeof(data));
				xstrncpy(data.name, ifr->ifr_name, sizeof(data.name));
			}
			sin = (struct sockaddr_in *) &ifr->ifr_addr;
			data.addr[data.n_addr++] = sin->sin_addr.s_addr;
		}
	}

	if (data.n_addr != 0)
		list_append(&data, (void **) int_list, sizeof(data));

	close(sock);
	return (0);
}

static int get_pcap_datalink(pcap_t *pd) {
	int datalink = pcap_datalink(pd);

	switch (datalink) {
		case DLT_RAW:
			return (0);
		case DLT_ATM_RFC1483:
			return (8);
		case DLT_EN10MB:
		case DLT_IEEE802:
			return (sizeof(struct ethhdr));
		case DLT_SLIP_BSDOS:
		case DLT_PPP_BSDOS:
			return (24);
		case DLT_SLIP:
			return (16);
		case DLT_PPP:
			return (4);
	}

	return (-1);
}

static int open_pcap_device(struct iff_data *idata, struct pcap_data *pdata) {
	u_char fstring[1024], nbuf[16], *temp;
	extern u_long flags;
	struct sockaddr_in sin;
	struct bpf_program filt;
	u_int i = 0;

	pdata->pd = pcap_open_live(idata->name, 1550, 0, 0, errbuf);

	if (pdata->pd == NULL)
		fatal("Fatal: pcap_open_live: %s", errbuf);

	pdata->dl = get_pcap_datalink(pdata->pd);

	if (pdata->dl == -1)
		return (-1);

	xstrncpy(fstring, "ip and (", sizeof(fstring));
	if (flags & LOG_TCP) {
		xstrncat(fstring, "tcp ", sizeof(fstring));
		++i;
	}
	if (flags & LOG_UDP) {
		if (i++ == 0)
			temp = "udp ";
		else
			temp = "or udp ";
		xstrncat(fstring, temp, sizeof(fstring));
	}
	if (flags & LOG_ICMP) {
		if (i == 0)
			temp = "icmp";
		else
			temp = "or icmp";
		xstrncat(fstring, temp, sizeof(fstring));
	}
	xstrncat(fstring, ") and dst host (", sizeof(fstring));

	for (i = 0 ; i < idata->n_addr ; i++) {
		sin.sin_addr.s_addr = idata->addr[i];
		if (i > 0)
			xstrncat(fstring, " or ", sizeof(fstring));
		inet_ntoa_r(&sin.sin_addr, nbuf, sizeof(nbuf));
		xstrncat(fstring, nbuf, sizeof(fstring));
	}

	xstrncat(fstring, ")", sizeof(fstring));

	if (pcap_compile(pdata->pd, &filt, fstring, 1, 0) == -1)
		fatal("Fatal: pcap_compile: %s", pcap_geterr(pdata->pd));

	if (pcap_setfilter(pdata->pd, &filt) == -1)
		fatal("Fatal: pcap_setfilter: %s", pcap_geterr(pdata->pd));

	return (0);
}

int setup_pcap(struct pcap_data **plist, const u_char *ifstring) {
	struct iff_data *ilist = NULL, *cur;
	struct pcap_data data;
	u_char *vif[32];

	if (ifstring != NULL) {
		u_int i = 0;
		u_char *delim, *pos;
		u_char *tstr = xstrdup(ifstring);

		pos = tstr;
		for (; i < 31 && (delim = strchr(pos, ',')) != NULL ; pos = delim) {
			*delim++ = '\0';
			vif[i++] = pos;
		}
		vif[i++] = pos;
		vif[i] = NULL;
		get_if_namesaddr(&ilist, (const u_char **) vif);
		free(tstr);
	} else
		get_if_namesaddr(&ilist, NULL);
	cur = ilist;
	while (cur != NULL) {
		open_pcap_device(cur, &data);
		list_append(&data, (void **) plist, sizeof(data));
		cur = cur->next;
	}

	list_destroy(ilist, NULL);
	return (0);
}

void parse_packet(struct ip *ip) {
	u_char buf[256];
	u_int h_len = ip->ip_hl << 2;
	u_int len = ntohs(ip->ip_len);
	bool found = false;
	struct frag_data *cur;
	struct ip *temp_ip = NULL;
	u_long hash = FRAGHASH(ip->ip_src.s_addr, ip->ip_dst.s_addr,
							ip->ip_id, ip->ip_p);

	/*
	** I think it's safe to assume most any TCP/IP implementation
	** will drop packets that fail either of the following two checks..
	*/
	if (h_len < sizeof(struct ip) || len < h_len) {
		mysyslog(LOG_INFO, "Warning: Short IP packet received from %s",
			host_lookup(&ip->ip_src, !(flags & NO_RESOLV), buf, sizeof(buf)));
		return;
	}
	
	if (in_cksum((u_short *) ip, h_len) != 0) {
		mysyslog(LOG_INFO,
			"Warning: IP packet with invalid checksum received from %s",
			host_lookup(&ip->ip_src, !(flags & NO_RESOLV), buf, sizeof(buf)));
		return;
	}

	pthread_mutex_lock(&frags[hash].lock);
	for (cur = frags[hash].head ; cur != NULL ; cur = cur->next) {
		if (cur->id == ip->ip_id &&
			cur->s_addr == ip->ip_src.s_addr &&
			cur->prot == ip->ip_p &&
			cur->d_addr == ip->ip_dst.s_addr)
		{
			found = true;
			break;
		}
	}
	pthread_mutex_unlock(&frags[hash].lock);
	
	if (ntohs(ip->ip_off) & ~IP_DF) {
		u_char lbuf[20];
		bool done = false;
		u_int offset = ntohs(ip->ip_off);
		struct ip_fragment tf, *temp;

		tf.off = (offset & IP_OFFMASK) << 3;
		tf.len = len - (ip->ip_hl << 2);

		if (tf.off + tf.len > 0xffff) {
			mysyslog(LOG_INFO,
				"Warning: Oversized fragment from %s",
				inet_ntoa_r(&ip->ip_src, lbuf, sizeof(lbuf)));
			if (cur != NULL) {
				pthread_mutex_lock(&cur->lock);
				deallocate_frags(cur, hash);
			}
			return;
		}

		if (found == false) {
			struct frag_data tfl;

			tfl.prot = ip->ip_p;
			tfl.id = ip->ip_id;
			tfl.s_addr = ip->ip_src.s_addr;
			tfl.d_addr = ip->ip_dst.s_addr;
			tfl.expire = 0;
			tfl.frag = NULL;

			tfl.bytes = 0;
			tfl.t_len = 0;
			tfl.rf = 0;
			
			/*
			** If we've reached the limit, get rid of the oldest. In practice,
			** this will hardly, if ever, happen, unless we're being attacked.
			*/
			if (frags[hash].count >= FRAG_MAX) {
				u_long oldt = -1UL;
				struct frag_data *q, *oldest = NULL;

				pthread_mutex_lock(&frags[hash].lock);
				for (q = frags[hash].head ; q != NULL ; q = q->next) {
					pthread_mutex_lock(&q->lock);
					if ((u_long) q->expire < oldt) {
						oldt = q->expire;
						oldest = q;
					}
					pthread_mutex_unlock(&q->lock);
				}
				pthread_mutex_unlock(&frags[hash].lock);
				pthread_mutex_lock(&oldest->lock);
				deallocate_frags(oldest, hash);
			}

			pthread_mutex_lock(&frags[hash].lock);
			cur = list_prepend(&tfl, (void **) &frags[hash].head, sizeof(tfl));
			pthread_mutex_init(&cur->lock, NULL);
			frags[hash].count++;
			pthread_mutex_unlock(&frags[hash].lock);
		}

		pthread_mutex_lock(&cur->lock);
		if (tf.off == 0) {
			memcpy(&cur->header, ip, sizeof(struct ip));
			cur->rf++;
		}

		if (!(offset & IP_MF)) {
			cur->t_len = tf.off + tf.len;
			cur->rf++;
		}

		if (found == false) {
			temp = list_prepend(&tf, (void **) &cur->frag, sizeof(tf));
			temp->data = xmalloc(tf.len);
			memcpy(temp->data, (char *) ip + h_len, tf.len);
		} else {
			struct ip_fragment *save;

			save = temp = cur->frag;

			while (temp != NULL) {
				if (tf.off < temp->off) {
					if ((temp->next && tf.off + tf.len > temp->next->off) ||
						(save != temp && tf.off < save->off + save->len))
					{
						mysyslog(LOG_INFO,
							"Warning: Overlapping IP fragments from %s",
							inet_ntoa_r(&ip->ip_src, lbuf, sizeof(lbuf)));
						deallocate_frags(cur, hash);
						return;
					} else {
						struct ip_fragment *ojnk = xmalloc(sizeof(tf));

						ojnk->off = tf.off;
						ojnk->len = tf.len;
						ojnk->data = xmalloc(ojnk->len);

						memcpy(ojnk->data, (char *) ip + h_len, ojnk->len);

						if (cur->frag == temp) {
							ojnk->next = temp;
							cur->frag = ojnk;
						} else {
							ojnk->next = temp;
							save->next = ojnk;
							temp->next = ojnk->next;
						}
						done = true;
					}
					break;
				} else {
					if (tf.off == temp->off) {
						mysyslog(LOG_INFO,
							"Warning: Duplicate IP fragments from %s",
							inet_ntoa_r(&ip->ip_src, lbuf, sizeof(lbuf)));
						deallocate_frags(cur, hash);
						return;
					} else {
						save = temp;
						temp = temp->next;
					}
				}
			}

			if (done == false) {
				if (tf.off < save->len + save->off) {
					mysyslog(LOG_INFO,
						"Warning: Overlapping IP fragments from %s",
						inet_ntoa_r(&ip->ip_src, lbuf, sizeof(lbuf)));
					deallocate_frags(cur, hash);
					return;
				}

				temp = list_append(&tf, (void **) &cur->frag, sizeof(tf));
				temp->data = xmalloc(tf.len);
				memcpy(temp->data, (char *) ip + sizeof(struct ip), tf.len);
			}
		}

		cur->expire = time(NULL) + FRAG_TTL;
		cur->bytes += tf.len;

		if (cur->rf < 2) {
			pthread_mutex_unlock(&cur->lock);
			return;
		}

		temp_ip = iplog_reassemble(cur);

		if (temp_ip == NULL) {
			pthread_mutex_unlock(&cur->lock);
			return;
		} else {
			ip = temp_ip;
			deallocate_frags(cur, hash);
		}
	} else if (cur != NULL) {
		pthread_mutex_lock(&cur->lock);
		deallocate_frags(cur, hash);
	}

	switch (ip->ip_p) {
		case IPPROTO_TCP:
			tcp_parser(ip);
			break;
		case IPPROTO_UDP:
			udp_parser(ip);
			break;
		case IPPROTO_ICMP:
			icmp_parser(ip);
			break;
	}

	if (temp_ip != NULL)
		free(temp_ip);
}

struct ip *iplog_reassemble(const struct frag_data *fl) {
	u_long off = sizeof(struct ip);
	size_t plen = sizeof(struct ip) + fl->bytes;
	struct ip *ip;
	struct ip_fragment *cur;

	if (fl->bytes != fl->t_len)
		return (NULL);

	/*
	** If we got here, we have all the fragments contained in a sorted, linked
	** list.  Just glue together all the pieces.
	*/
	ip = xmalloc(plen);
	memcpy(ip, &fl->header, off);

	for (cur = fl->frag ; cur != NULL ; cur = cur->next)
		memcpy((char *) ip + (cur->off + off), cur->data, cur->len);

	ip->ip_len = htons(plen);
	/* We've ignored IP options, if any were present.. */
	ip->ip_hl = (sizeof(struct ip)) >> 2;
	
	/* Don't bother generating a new checksum, as we'll never use it */
	return (ip);
}

/* This must always be called with fl->lock held. */
void deallocate_frags(struct frag_data *fl, u_long hash) {
	list_destroy(fl->frag, frag_cleanup);
	pthread_mutex_lock(&frags[hash].lock);
	list_delete(fl, (void **) &frags[hash].head);
	--frags[hash].count;
	pthread_mutex_unlock(&frags[hash].lock);
}

void frag_cleanup(void *data) {
	struct ip_fragment *frag = data;

	free(frag->data);
}

void expire_frags(void) {
	struct frag_data *cur;
	u_int i;
	
	for (i = 0 ; i < FRAG_TSIZE ; i++) {
		pthread_mutex_lock(&frags[i].lock);

		for (cur = frags[i].head ; cur != NULL ;) {
			pthread_mutex_lock(&cur->lock);
			if (cur->expire >= time(NULL)) {
				list_destroy(cur->frag, frag_cleanup);
				cur = list_delete(cur, (void **) &frags[i]);
				--frags[i].count;
			} else {
				pthread_mutex_unlock(&cur->lock);
				cur = cur->next;
			}
		}

		pthread_mutex_unlock(&frags[i].lock);
	}
}

#ifdef HAVE_PTHREAD_CANCEL
void destroy_frag_table(void) {
	struct frag_data *cur;
	u_int i;

	/*
	** We've just canceled all the threads, and the locks could be in any
	** state, but it really doesn't matter now.  Just destroy the table and
	** reset everything.
	*/
	for (i = 0 ; i < FRAG_TSIZE ; i++) {
		for (cur = frags[i].head ; cur != NULL ; cur = cur->next)
			list_destroy(cur->frag, frag_cleanup);
		list_destroy(frags[i].head, NULL);
		frags[i].head = NULL;
		frags[i].count = 0;
		pthread_mutex_init(&frags[i].lock, NULL);
	}
}
#endif

/* vim:ts=4:sw=8:tw=0 */
