/*
** iplog_tcp.c - iplog TCP traffic logger.
** Copyright (C) 1999 Odin <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
*/

#include <config.h>

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <pthread.h>
#include <errno.h>

#include <iplog.h>
#include <iplog_options.h>
#include <iplog_filter.h>
#define DEF_ONLY
#include <iplog_scan.h>

#define TH_BOG		(~(TH_FIN | TH_SYN | TH_RST | TH_PUSH | TH_ACK | TH_URG))
#define SYNFLOOD	20

extern u_long flags;

static time_t synlast;
static u_long last_bogus;
static u_char syncount;
static int raw_sock;

pthread_mutex_t syn_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t bogus_lock = PTHREAD_MUTEX_INITIALIZER;

bool get_raw_sock(void) {
	raw_sock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);

	return (raw_sock != -1);
}

__inline__ bool tcp_res(void) {
	return ((flags & TCP_RES) && !(flags & NO_RESOLV) && syncount != SYNFLOOD);
}

int tcp_parser(struct ip *ip) {
	struct tcphdr *tcp = (struct tcphdr *) ((char *) ip + (ip->ip_hl << 2));
	time_t now = time(NULL);
	u_long len = ntohs(ip->ip_len) - (ip->ip_hl << 2);
	u_char lbuf[MAX_HSTLEN];
	u_int tflags = tcp->th_flags;

	if (apply_filter(FIL_TCP, ip, tcp->th_sport, tcp->th_dport))
		return (0);

	/*
	** This seems to be enough to confuse programs like nmap and queso
	** that try to determine the operating system by sending bogus
	** packets.  As a side effect, it causes several of nmap's "stealth"
	** scans not to work.
	*/

	if ((flags & FOOL_NMAP) &&
		((tflags & TH_BOG) || (tflags == TH_PUSH) || (tflags == 0) ||
		((tflags & TH_SYN) && (tflags & (TH_FIN | TH_RST)))))
	{
		u_char *temp = xcalloc(1, sizeof(struct ip) + sizeof(struct tcphdr));
		struct ip *xip;
		struct tcphdr *xtcp;
		struct sockaddr_in sin;

		xip = (struct ip *) temp;
		xtcp = (struct tcphdr *) ((u_char *) temp + sizeof(struct ip));

		xip->ip_hl = sizeof(struct ip) >> 2;
		xip->ip_v = 4;
		xip->ip_len = htons(sizeof(struct ip) + sizeof(struct tcphdr));
		xip->ip_tos = ip->ip_tos;
		xip->ip_id = rand();
		xip->ip_off = htons(rand() & 1 ? IP_DF : 0);
		xip->ip_ttl = rand();
		xip->ip_p = ip->ip_p;
		xip->ip_dst.s_addr = ip->ip_src.s_addr;
		xip->ip_src.s_addr = ip->ip_dst.s_addr;
		xip->ip_sum = in_cksum((u_short *) xip, sizeof(struct ip));
		
		xtcp->th_sport = tcp->th_dport;
		xtcp->th_dport = tcp->th_sport;
		xtcp->th_seq = tcp->th_seq;
		xtcp->th_ack = tcp->th_ack;
		xtcp->th_x2 = tcp->th_x2;
		xtcp->th_off = rand();
		xtcp->th_flags = rand();
		xtcp->th_win = rand() & 1;
		xtcp->th_urp = rand();

		xtcp->th_sum = in_cksum((u_short *) xtcp, sizeof(struct tcphdr));

		memset(&sin, 0, sizeof(sin));
		sin.sin_family = AF_INET;
		sin.sin_port = tcp->th_sport;
		sin.sin_addr.s_addr = ip->ip_src.s_addr;

		sendto(raw_sock, xip, sizeof(struct ip) + sizeof(struct tcphdr),
				0, &sin, sizeof(struct sockaddr_in));

		free(temp);
	}

	if (flags & SYN_FLOOD) {
		pthread_mutex_lock(&syn_lock);
		if (syncount > 0) {
			syncount -= (now - synlast) << 2;
			synlast = now;
		}
		pthread_mutex_unlock(&syn_lock);
	}
	
	if ((flags & NULL_SCAN) && !tcp->th_flags &&
		check_scan(SCAN_NULL, ip->ip_src.s_addr, len) != 0)
	{
		return (0);
	}

	if ((flags & XMAS_SCAN) && (tcp->th_flags & TH_FIN) &&
		(tcp->th_flags & TH_URG) && (tcp->th_flags & TH_PUSH) &&
		check_scan(SCAN_XMAS, ip->ip_src.s_addr, len) != 0)
	{
		return (0);
	}

	if ((flags & FIN_SCAN) && (tcp->th_flags & TH_FIN) &&
		!(tcp->th_flags & TH_ACK) &&
		check_scan(SCAN_FIN, ip->ip_src.s_addr, len) != 0)
	{
		return (0);
	}

	if ((flags & BOGUS) && (tcp->th_flags & TH_BOG)) {
		pthread_mutex_lock(&bogus_lock);
		if (last_bogus != ip->ip_src.s_addr) {
			last_bogus = ip->ip_src.s_addr;
			pthread_mutex_unlock(&bogus_lock);
			mysyslog(PRIORITY, "TCP: Bogus TCP flags set by %s",
				host_lookup(&ip->ip_src, tcp_res(),	lbuf, sizeof(lbuf)));
		} else
			pthread_mutex_unlock(&bogus_lock);

		return (0);
	}

	if ((tcp->th_flags & TH_SYN) && !(tcp->th_flags & TH_ACK)) {
		if ((flags & PORTSCAN) &&
			check_scan(SCAN_PORT, ip->ip_src.s_addr, len) != 0)
		{
				return (0);
		}

		pthread_mutex_lock(&syn_lock);
		if (syncount != SYN_FLOOD)
			++syncount;
		pthread_mutex_unlock(&syn_lock);

		if ((flags & GET_IDENT) && is_listening(tcp->th_dport)) {
			pthread_t pt;
			u_long buflen = ntohs(ip->ip_len);
			u_char *buf;

			buf = xmalloc(buflen);
			memcpy(buf, ip, buflen);
			pthread_create(&pt, NULL, get_ident_data, buf);
			pthread_detach(pt);

			return (0);
		} else {
			u_char sbuf[MAX_SRVLEN];

			mysyslog(PRIORITY, "TCP: %s connection attempt from %s:%u",
				serv_lookup(tcp->th_dport, "tcp", sbuf, sizeof(sbuf)),
				host_lookup(&ip->ip_src, tcp_res(), lbuf, sizeof(lbuf)),
				ntohs(tcp->th_sport));
		}
	}

	return (0);
}

/* vim:ts=4:sw=8:tw=0 */
